import numpy as np
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))
from jgtutils.jgtconstants import VOLUME
from mlconstants import TARGET
from mlconstants import MFI_DEFAULT_COLNAME,ZONE_DEFAULT_COLNAME
import mfihelper2 as mfihelper
from mfihelper2 import get_mfi_features_column_list_by_timeframe
import anhelper
from mxconstants import readmx_drop_columns_arr

import jtc
import pandas as pd

def _read_mx_and_prep_02(i,t,drop_columns_arr = None,dropna_volume=True):
 
    
  df=jtc.readMXFile(instrument=i,timeframe=t)
  try:
    if drop_columns_arr is not None:
      for col in drop_columns_arr:
        if col in df.columns:
          df.drop(columns=[col],inplace=True)    
  except:
    pass
  #drop rows with Volume=0
  if dropna_volume:
    df=df[df[VOLUME]!=0]
  return df





# utility
def mk_safename_namespace_path(i,t,x_fn_namespace,sub_namespace,suffix_base="",out_dir="",bs=""):
  ifn=i.replace('/','-')
  bs = "_sell" if bs=="S" else "_buy" if bs=="B" else ""

  fn=f"{x_fn_namespace}_{sub_namespace}_{ifn}_{t}{suffix_base}{bs}.csv"
  if out_dir!="":
    fn=os.path.join(out_dir,fn)
    #make sure the directory exists   
    os.makedirs(os.path.dirname(fn), exist_ok=True)
  return fn


from mxconstants import mx_common_columns
def _extract_mfi_structure_dataframe(mxdf,t,common_columns = mx_common_columns):
  
  mfi_feature_columns_list = get_mfi_features_column_list_by_timeframe(t)
  print("Extracted MFI Features columns list:",mfi_feature_columns_list)
  combined_columns = mfi_feature_columns_list+common_columns
  df_with_features=mxdf[combined_columns]
  return df_with_features

#@STCIssue Location could move to a Common to MX and Reality analysis (mfihelper??)
def _add_lag_features_to_dataframe(df: pd.DataFrame, t, lag_period=1, total_lagging_periods=5,out_lag_midfix_str='_lag_',inplace=True):
  if not inplace:
    df = df.copy()
  columns_to_add_lags_to = get_mfi_features_column_list_by_timeframe(t, MFI_DEFAULT_COLNAME)
  #columns_to_add_lags_to.append(MFI_DEFAULT_COLNAME) #We want a lag for the current TF
  
  anhelper.add_lagging_columns(df, columns_to_add_lags_to, lag_period, total_lagging_periods, out_lag_midfix_str)
  #convert columns_to_add_lags_to to type int
  # for col in columns_to_add_lags_to: #@STCIssue Isn't that done already ???  Or it thinks they are Double !!!!
  #   for j in range(1, total_lagging_periods + 1):
  #     df[f'{col}{out_lag_midfix_str}{j}']=df[f'{col}{out_lag_midfix_str}{j}'].astype(int)
  
  return df




  
def wf_get_mxdf_and_add_mfi_features_to_df(i,t,common_columns = [MFI_DEFAULT_COLNAME,TARGET, 'vaoc','fdb'],drop_columns_arr = None)->pd.DataFrame:
  if drop_columns_arr is None:
    drop_columns_arr = readmx_drop_columns_arr
    
  df=_read_mx_and_prep_02(i,t,drop_columns_arr)
  mfihelper.column_mfi_str_in_dataframe_to_id(df,t) #@STCIssue WE WONT NEED THIS when using MFI_SIGNAL instead of MFI_VAL
  _extract_mfi_structure_dataframe(df,t,common_columns)
  #add lag
  _add_lag_features_to_dataframe(df,t)
  return df




def _select_where_target_is_not_zero(df, target_colname='target'):
  dfresult=df[df[target_colname]!=0].copy()
  return dfresult


def get_analysis_data_240702(i,t,bs,target_colname='target',signal_column='fdb',drop_signal_column=True,signal_column_sell_value=-1,signal_column_buy_value=1,quiet=False):
  """
  Get the analysis data for the prototype. This is a wrapper function that calls the other functions in this module. 
  """
  df=wf_get_mxdf_and_add_mfi_features_to_df(i,t)
  df=_select_where_target_is_not_zero(df,target_colname)
  if bs=='S' or bs=='s' or bs=='sell' or bs=='SELL' or bs=='Sell':
    print('Selecting sell signals') if not quiet else None
    df=df[df[signal_column]==signal_column_sell_value]
  else:
    df=df[df[signal_column]==signal_column_buy_value]
    print('Selecting buy signals') if not quiet else None
  if drop_signal_column:
    df.drop(columns=[signal_column],inplace=True)
  return df

def _drop_column_part01(df,columnsToDrop = ['vaos','vaob','vaosc','vaobc','fh8', 'fl8', 'fh89', 'fl89','mfi','aoaz','aobz','sz','bz','acb','acs','ss','sb','mfi_sq', 'mfi_green','mfi_fade', 'mfi_fake','tmax', 'tmin', 'p', 'l'],inplace=True):
  """
  Drop columns from the dataframe part 01 of our prototype
  
  Parameters:
  df: pd.DataFrame - the dataframe to drop columns from
  columnsToDrop: list - the list of columns to drop
  
  Returns:
  pd.DataFrame - the dataframe with the columns dropped
  """
  if not inplace:
    df = df.copy()
  for col in columnsToDrop:
    if col in df.columns:
      df.drop(columns=[col],inplace=True)
  return df

def _drop_column_part02(df,more2dropcolumns=['Volume', 'Open', 'High', 'Low', 'Close', 'Median', 'ac', 'mfi_sig', 'zcol_M1', 'zcol_W1', 'ao_W1', 'vaoc','zlc', 'zlcb', 'zlcs', 'zcol','fh', 'fl', 'fh3', 'fl3', 'fh5', 'fl5', 'fdbb', 'fdbs','jaw', 'teeth', 'lips', 'bjaw', 'bteeth', 'blips', 'tjaw','tteeth', 'tlips','zcol_D1'],inplace=True):
  """
  Drop columns from the dataframe part 02 of our prototype
  
  Parameters:
  df: pd.DataFrame - the dataframe to drop columns from
  more2dropcolumns: list - the list of columns to drop
  
  Returns:
  pd.DataFrame - the dataframe with the columns dropped
  """
  if not inplace:
    df = df.copy()
  for col in more2dropcolumns:
    if col in df.columns:
      df.drop(columns=[col],inplace=True)
  return df

def get_analysis_data_240702_cleaned(i,t,bs,target_colname='target',signal_column='fdb',drop_signal_column=True):
  """Get the analysis data for the prototype. This is a wrapper function that calls the other functions in this module. It also drops columns that are not needed for the prototype.
  """
  df=get_analysis_data_240702(i,t,bs,target_colname,signal_column,drop_signal_column)
  df=_drop_column_part01(df)
  df=_drop_column_part02(df)
  return df






# This file is generated by pto-pre-build.sh. Do not edit it manually.







# DONT EDIT THIS FILE MANUALLY. IT IS GENERATED BY pto-pre-build.sh








#DONT EDIT THIS FILE MANUALLY. IT IS GENERATED BY pto-pre-build.sh






# Mon 19 Aug 2024 05:12:58 AM EDT
# SOURCE NAME: /b/Dropbox/jgt/drop/fnml.py
########################
 

import argparse
import json
import os
import subprocess
import sys

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))
from jgtutils import jgtcommon
from jgtutils.jgtcliconstants import (CLI_FXADDORDER_PROG_NAME,CLI_FXMVSTOP_PROG_NAME,CLI_FXRMORDER_PROG_NAME,CLI_FXRMTRADE_PROG_NAME,CLI_FXTR_PROG_NAME,PDSCLI_PROG_NAME)

from jgtutils.jgtconstants import (LIPS,TEETH,JAW)
from jgtutils.jgtconstants import FDB,HIGH,LOW,CLOSE

from jgtpy.jgtpyconstants import (IDSCLI_PROG_NAME,CDSCLI_PROG_NAME,ADSCLI_PROG_NAME,MKSCLI_PROG_NAME,JGTCLI_PROG_NAME)

from jgtutils.FXTransact import FXTransactWrapper,FXTransactDataHelper as ftdh,FXTrades,FXTrade



from jgtpy.JGTIDS import _ids_add_fdb_column_logics_v2
from jgtpy import jgtapyhelper as th

#from jgtpy.JGTIDSSvc import get_ids
#from jgtpy.JGTIDSRequest import JGTIDSRequest

from SOHelper import get_bar_at_index,get_last_two_bars

from mlcliconstants import (MLFCLI_PROG_NAME,TTFCLI_PROG_NAME,PNCLI_PROG_NAME,MXCLI_PROG_NAME)

TFW_PROG_NAME = "tfw"

def w(timeframe,script_to_run=None,exit_on_timeframe=False):
  if script_to_run:
    subprocess.run([TFW_PROG_NAME, '-t', timeframe,'-N', '-B', script_to_run], check=True)
  elif exit_on_timeframe:
    subprocess.run([TFW_PROG_NAME, '-t', timeframe,'-N', '-X'], check=True)
    exit()
  else:
    subprocess.run([TFW_PROG_NAME, '-t', timeframe,'-N'], check=True)



def fxaddorder( instrument, lots, rate, buysell, stop, demo=False,flag_pips=False):
  pips_arg = '--pips' if flag_pips else ''
  demo_arg = '--demo' if demo else '--real'
  subprocess.run([CLI_FXADDORDER_PROG_NAME, '-i', instrument, '-n', lots, '-r', rate, '-d', buysell, '-x',stop,pips_arg , demo_arg], check=True)

def _get_instrument_from_orderid(orderid):
  #get the instrument from the orderid
  #fxtr -id 68782480 --demo
  #get the instrument from the orderid
  raise NotImplementedError("get the instrument from the orderid")

def _get_buysell_from_orderid(orderid):
  #get the buysell from the orderid
  #fxtr -id 68782480 --demo
  #get the buysell from the orderid
  raise NotImplementedError("get the buysell from the orderid")

def _get_stop_rate_from_orderid(orderid):
  #get the stop rate from the orderid
  #fxtr -id 68782480 --demo
  #get the stop rate from the orderid
  raise NotImplementedError("get the stop rate from the orderid")

def entryvalidate(orderid,timeframe, demo=False):
  demo_arg = '--demo' if demo else '--real'
  instrument=_get_instrument_from_orderid(orderid)
  bs=_get_buysell_from_orderid(orderid)
  stop_rate=_get_stop_rate_from_orderid(orderid)
  df = _get_ids_updated(instrument, timeframe)
  cb=get_bar_at_index(df,-1)
  clow=cb[LOW]
  cclose=cb[CLOSE]
  chigh=cb[HIGH]
  if bs=="B" and (cclose<stop_rate or clow<stop_rate):
    print("The stop rate has hit")
    fxrmorder(orderid, demo=demo)
  else:
    if bs=="S" and (cclose>stop_rate or chigh>stop_rate):
      print("The stop rate has hit")
      fxrmorder(orderid, demo=demo)
  
  
def fxrmorder(orderid, demo=False):
  demo_arg = '--demo' if demo else '--real'
  subprocess.run([CLI_FXRMORDER_PROG_NAME, '-id', orderid, demo_arg], check=True)

def fxrmtrade(tradeid, demo=False):
  demo_arg = '--demo' if demo else '--real'
  subprocess.run([CLI_FXRMTRADE_PROG_NAME, '-tid', tradeid, demo_arg], check=True)

def fxtr(tradeid=None,orderid=None, demo=False,save_flag=True):
  save_arg = '-save' if save_flag else ''
  demo_arg = '--demo' if demo else '--real'
  if tradeid:
    subprocess.run([CLI_FXTR_PROG_NAME, '-tid', tradeid, demo_arg, save_arg])
  elif orderid:
    subprocess.run([CLI_FXTR_PROG_NAME, '-id', orderid, demo_arg, save_arg])
  else:
    subprocess.run([CLI_FXTR_PROG_NAME, demo_arg, save_arg])
  msg = "File saved."
  print_jsonl_message(msg,extra_dict={"trade_id":tradeid,"order_id":orderid },scope="jgtapp::fxtr")

def fxmvstop(tradeid,stop,flag_pips=False, demo=False):
  pips_arg = '--pips' if flag_pips else ''
  demo_arg = '--demo' if demo else '--real'
  cli_args = [CLI_FXMVSTOP_PROG_NAME, '-tid', tradeid, '-x', stop, demo_arg]
  if pips_arg != '':
    cli_args.append(pips_arg)
  subprocess.run(cli_args)

def ids(instrument, timeframe,use_full=False,use_fresh=True):
  use_fresh_arg = '-old' if not use_fresh else '--fresh'
  use_full_arg = '--full' if use_full else '-new'
  subprocess.run([IDSCLI_PROG_NAME, '-i', instrument, '-t', timeframe, use_full_arg, use_fresh_arg])

"""
fxmvstopgator -tid 68773276  --demo -i AUD/NZD -t H4 --lips
"""
#@STCGoal RM Order if Stop was Hit





#@STCGoal Move EXIT Stop On FDB Signal
fxmvstopfdb_epilog = "Move the stop to the FDB signal. If the stop is already hit, close the trade if --close is passed.  Considering we could add --lips to move the stop to the lips if no FDB signal is found #@STCIssue It would requires to detect if previously the stop was moved to an FDB signal (as we dont want to move it again back in the lips if we are still in an FDB signal)"

MOVED_TO_FDB_FLAG_NAME = "moved_to_fdb"
def fxmvstopfdb(i,t,tradeid,demo=False,close=False,lips=False,teeth=False,jaw=False,not_if_stop_closer=True):
  demo_arg="--demo" if demo else "--real"
  if close:
    print("Closing the trade if the stop of fdbsignal is already hit")
    raise NotImplementedError("Closing the trade if the stop of fdbsignal is already hit")
  
  
  
  
  #Update the local datafor the trade
  ## we expect : fxtransact_68782480.json
  #fxtr(tradeid=tradeid,demo=demo)
  
  trade_data = _get_trade_data(tradeid, demo)
  #just make sure we have the same instrument
  if i == "_": # use the instrument from the trade data 
    i = trade_data["instrument"]
  if not trade_data["instrument"]==i:
    print(f"Trade data instrument {trade_data['instrument']} is different from the passed instrument {i}")
    from jgtutils.jgterrorcodes import INSTRUMENT_NOT_VALID_EXIT_ERROR_CODE
    sys.exit(INSTRUMENT_NOT_VALID_EXIT_ERROR_CODE)
  

  #fuck=fxtrades[tradeid]
  #get the trade data
  #print(fxdata)
  #trade_data=ftdh.load_fxtrade_from_fxtransact(fxdata,tradeid)
  #trade_data2=fxdata.get_trade(tradeid)
  direction=trade_data["buy_sell"]
  original_stop=trade_data["stop"]
  msg = "Detected trade direction."
  print_jsonl_message(msg,extra_dict={"direction":direction,"original_stop":original_stop})
  # print(trade_data)
  #@STCIssue Why do I want that info on the Trade ??
  
    
  skip_generating_ids=False
  #if SKIP_IDS defined in os env, skip
  if os.getenv("SKIP_IDS"):
    skip_generating_ids=True
    
  if skip_generating_ids:
    print("Skipping generating IDS (JUST READING  IT FOR DEV)")
  

  dfc = _get_ids_updated(i, t)
  lcb=get_bar_at_index(dfc,-2)
  lcbfdb = lcb[FDB]
  lcbhigh=lcb[HIGH]
  lcblow=lcb[LOW]
  cb=get_bar_at_index(dfc,-1)
  clow=cb[LOW]
  cclose=cb[CLOSE]
  chigh=cb[HIGH]
  
  is_sell_fdb=lcbfdb<0
  is_buy_fdb=lcbfdb>0
  stop_price=None
  if direction=="B" and is_sell_fdb:
    stop_price=lcblow
  else:
    if direction=="S" and is_buy_fdb:
      stop_price=lcbhigh
      
  
  if stop_price:
    msg_ = f"We change the Stop for:{stop_price} from : {original_stop}"
    print(msg_)
    #@q Does our Current Bar Broke the FDB Signal ??
    #@STCIssue If close is True, we should close the trade if the stop is already hit
    stop_has_hit=False
    if direction=="B" and cclose>stop_price:
      stop_has_hit=True
    else: 
      if direction=="S" and cclose<stop_price:
        stop_has_hit=True
    msg=f"Stop has hit" if stop_has_hit else "Stop has not hit"
    print(msg)
    
    if stop_has_hit:
      msg = "The stop has hit the FDB signal, we are closing the trade"
      print_jsonl_message(msg,scope="fxmvstopfdb")
      _close_trade_cmd=f"jgtapp fxrmtrade -tid {tradeid} {demo_arg}"
      msg = f"Executing:{_close_trade_cmd}"
      print_jsonl_message(msg)
      fxrmtrade(tradeid,demo=demo)
      #print("Done")
    else:
      print("We are moving the stop to the FDB signal")
      fxmvstop(tradeid,stop_price,demo=demo)
      #write a flag so we know we moved to an FDB Signal
      flag_object={MOVED_TO_FDB_FLAG_NAME:True}
      flag_file_path = __get_fdb_moved_flag_filename(tradeid)
      with open(flag_file_path,"w"):
        json.dump(flag_object)
  
      
  else:
    #msg = "We dont have a signal, Shall we fall in set the stop to the lips ??"
    #print_jsonl_message(msg,scope="fxmvstopfdb")
    if lips or teeth or jaw:
      
      flag_moved_to_fdb=_check_flag_fdb_moved(tradeid)
      
      if flag_moved_to_fdb:
        msg = "We already moved to an FDB signal, we are not moving the stop back to the lips."
        print_jsonl_message(msg,scope="fxmvstopfdb",extra_dict={"trade_id":tradeid, "goal":"Know we had an FDB signal.  Could that be invalidated by the current bar broking the other side of our FDB Signal, therefore we would come back to the probably --lips"})
      else:
          
        line_arg="lips" if lips else "teeth" if teeth else "jaw"
        fxmvstopgator_cmd = f"jgtapp fxmvstopgator -i {i} -t {t} -tid {tradeid} --{line_arg} {demo_arg}"
        #print("We are running something like this:",fxmvstopgator_cmd)
        msg = f"Moving stop to {line_arg}"
        print_jsonl_message(msg,extra_dict={"trade_id":tradeid,"instrument":i,"timeframe":t, "line":line_arg})
        fxmvstopgator(i,t,tradeid,lips=lips,teeth=teeth,jaw=jaw,demo=demo,skip_trade_data_update=True)
    else:
      msg = "No --lips, --teeth or --jaw passed, we are not moving the stop"
      print_jsonl_message(msg)

def _get_trade_data(tradeid, demo,fresh=True):
    fxtr(demo=demo,tradeid=tradeid)
    expected_fn=f"fxtransact_{tradeid}.json"
    from jgtutils.jgtfxhelper import mkfn_cfxdata_filepath
    expected_path=mkfn_cfxdata_filepath(expected_fn)
    #expected_path=os.path.join("data","jgt",expected_fn)
    #x=FXTrade.from_id(tradeid)
    #expected_path=ftdh.load_fxtransact_from_file(expected_fn)
    if not os.path.exists(expected_path):
      fresh=True
      print(f"File {expected_path} does not exist - fxtr will be ran")
    else:
      fresh=False # We have the file, we dont need to run fxtr
    
    if fresh:
      # fxtr(tradeid=tradeid,demo=demo)
      fxtr(demo=demo)
    #return
    msg = "Reading the trade data from the file:"
    print_jsonl_message(msg,extra_dict={"trade_id":tradeid,"demo":demo, "expected_fn":expected_path},scope="jgtapp")
    fxdata=ftdh.load_fxtransact_from_file(expected_path)
    trade_data:FXTrade=None
    trades:FXTrades=fxdata.trades
    #return None if no trades in the list
    for trade in trades.trades:
    #print(trade)
      if str(trade["trade_id"])==str(tradeid):
        trade_data=trade
        break
    return trade_data

def _check_flag_fdb_moved(tradeid):
    flag_file_path = __get_fdb_moved_flag_filename(tradeid)
    flag_moved_to_fdb=False
    if os.path.exists(flag_file_path):
        #Load it and check if we moved to an FDB signal is set to true
      with open(flag_file_path,"r") as f:
        flag_object=json.load(f)
        if MOVED_TO_FDB_FLAG_NAME in flag_object:
          flag_moved_to_fdb=flag_object[MOVED_TO_FDB_FLAG_NAME]
    return flag_moved_to_fdb

def __get_fdb_moved_flag_filename(tradeid):
    return f".jgt/fdb_moved_flag_{tradeid}.json"
    
  
  
  

def _get_ids_updated(i, t,skip_generating=False):
    try:
      if not skip_generating:
        ids(i,t,use_fresh=True,use_full=False)
    except:
      print("IDS failed")
  #read the last bar from the IDS csv
    df=th.read_ids(i,t)
    dfc=_ids_add_fdb_column_logics_v2(df)
    return dfc
    
  
def print_jsonl_message(msg,extra_dict:dict=None,scope=None):
  o={}
  o["message"]=msg
  if extra_dict:
      o.update(extra_dict)
  if scope:
    o["scope"]=scope
  print(json.dumps(o))

def fxmvstopgator(i,t,tradeid,lips=True,teeth=False,jaw=False,demo=False,skip_trade_data_update=False,loop_action=False):
  

  trade_data=_get_trade_data(tradeid, demo,fresh=not skip_trade_data_update)
  if trade_data is None or trade_data=={}:
    msg = f"Trade data not found for tradeid {tradeid}"
    print_jsonl_message(msg)
    from jgtutils.jgterrorcodes import TRADE_NOT_FOUND_EXIT_ERROR_CODE
    sys.exit(TRADE_NOT_FOUND_EXIT_ERROR_CODE)
    
  if i == "_": # use the instrument from the trade data
    i = trade_data["instrument"]
  
  #First update the IDS
  df = _get_ids_updated(i, t)
  #get the last bar
  current_bar=get_bar_at_index(df,-1)
  choosen_line="_"
  #get the stop from choosen indicator line (by flag)
  if not teeth and not jaw:
    lips=True # DEFAULT
  if lips:
    stop=str(current_bar[LIPS])
    choosen_line="lips"
  elif teeth:
    stop=str(current_bar[TEETH])
    choosen_line="teeth"
  elif jaw:
    stop=str(current_bar[JAW])
    choosen_line="jaw"
  else:
    raise ValueError("No indicator line selected")
  msg = f"Made Gator line stop."
  print_jsonl_message(msg,extra_dict={"stop":stop,"line":choosen_line},scope="fxmvstopgator")
  #Then move the stop
  fxmvstop(tradeid,stop,demo=demo)
  
  if loop_action:
    w(t,exit_on_timeframe=True)
    #recursive call
    fxmvstopgator(i,t,tradeid,lips=lips,teeth=teeth,jaw=jaw,demo=demo,skip_trade_data_update=True,loop_action=True)

def tide(instrument, timeframe, buysell):
  raise DeprecationWarning("tide is deprecated. ")
  #subprocess.run(['tide', instrument, timeframe, buysell])

def pds(instrument, timeframe,use_full=True):
  use_full_arg = '--full' if use_full else ''
  subprocess.run([PDSCLI_PROG_NAME, '-i', instrument, '-t', timeframe, use_full_arg])

def cds(instrument, timeframe, use_fresh=False,use_full=True):
  use_full_arg = '--full' if use_full else ''
  old_or_fresh = '-old' if not use_fresh else '--fresh'
  subprocess.run([CDSCLI_PROG_NAME, '-i', instrument, '-t', timeframe,use_full_arg, old_or_fresh])

def ads(instrument, timeframe, use_fresh=False,tc=True,pov=False):
  old_or_fresh = '-old' if not use_fresh else '--fresh'
  ads_cli_args = [ADSCLI_PROG_NAME, '-i', instrument, '-t', timeframe, old_or_fresh]
  if tc and not pov:
    ads_cli_args.append('-sf')
    ads_cli_args.append('tc')
  elif pov:
    ads_cli_args.append('-sf')
    ads_cli_args.append('cp')
  subprocess.run(ads_cli_args)

def ocds(instrument, timeframe,use_full=True):
  use_full_arg = '--full' if use_full else ''
  subprocess.run([JGTCLI_PROG_NAME, '-i', instrument, '-t', timeframe, use_full_arg, '-old'])

  
def ttf(instrument, timeframe,pn="ttf",use_fresh=False,use_full=True):
  use_full_arg = '--full' if use_full else ''
  use_fresh_arg = '-old' if not use_fresh else '--fresh'
  
  ttf_args = [TTFCLI_PROG_NAME, '-i', instrument, '-t', timeframe, use_full_arg, use_fresh_arg, '-pn', pn]
  print("TTF is being ran by jgtapp with args: ", ttf_args)
  subprocess.run(ttf_args)


def mlf(instrument, timeframe,pn="ttf",total_lagging_periods=5,use_fresh=False,use_full=True):
  use_full_arg = '--full' if use_full else ''
  use_fresh_arg = '-old' if not use_fresh else '--fresh'
  mlf_args = [MLFCLI_PROG_NAME, '-i', instrument, '-t', timeframe, use_full_arg, use_fresh_arg, '-pn', pn,'--total_lagging_periods',total_lagging_periods]
  subprocess.run(mlf_args)

  

def mx(instrument, timeframe, use_fresh=False):
  old_or_fresh = '-old' if not use_fresh else '--fresh'
  subprocess.run([MXCLI_PROG_NAME, '-i', instrument, '-t', timeframe, old_or_fresh])

def ttfmxwf(instrument, use_fresh=False):
  for t in ["M1", "W1", "D1", "H4"]:
    print(f"Processing {instrument} for timeframe {t}")
    print("  CDS....")
    cds(instrument, t, use_fresh)
    if t != "M1":
      print("  TTF....")
      ttf(instrument, t)
    if t != "M1" and t != "W1":
      print("  MX....")
      mx(instrument, t)
      
def ttfwf(instrument, use_fresh=False):
  
  _settings = jgtcommon.get_settings()
  #ttf2run
  if hasattr(_settings, 'ttf2run'):
    ttf2run=_settings.ttf2run
  else:
    from jgtml.mlconstants import TTF2RUN_DEFAULT
    ttf2run=TTF2RUN_DEFAULT
  for pn in ttf2run:
      
    for t in ["M1", "W1", "D1", "H4"]:
      print(f"Processing {instrument} for timeframe {t}")
      print("  CDS....")
      cds(instrument, t, use_fresh)
      if t != "M1":
        print("  TTF....")
        ttf(instrument, t,pn=pn)
        
def main():
  cli_help_description = "CLI equivalent of bash functions"
  parser = jgtcommon.new_parser(cli_help_description,"JGTApp Wrapper/runner - run various ","jgtapp",add_exiting_quietly_flag=True)#argparse.ArgumentParser(description=cli_help_description)
  subparsers = parser.add_subparsers(dest='command')
  
  #fxaddorder
  parser_fxaddorder = subparsers.add_parser('fxaddorder', help='Add an order')
  parser_fxaddorder.add_argument('-i','--instrument', help='Instrument')
  parser_fxaddorder.add_argument('-n','--lots', help='Lots')
  parser_fxaddorder.add_argument('-r','--rate', help='Rate')
  parser_fxaddorder.add_argument('-d','--buysell', help='Buy or Sell')
  parser_fxaddorder.add_argument('-x','--stop', help='Stop')
  parser_fxaddorder.add_argument('--demo', action='store_true', help='Use the demo account')
  parser_fxaddorder.add_argument('--real', action='store_true', help='Use the real account',default=True)
  parser_fxaddorder.add_argument('--pips', action='store_true', help='Use pips')
  
  #fxrmorder
  parser_fxrmorder = subparsers.add_parser('fxrmorder', help='Remove an order')
  parser_fxrmorder.add_argument('-id','--orderid', help='Order ID')
  parser_fxrmorder.add_argument('--demo', action='store_true', help='Use the demo account')
  parser_fxrmorder.add_argument('--real', action='store_true', help='Use the real account',default=True)
  
  #entryvalidate
  parser_entryvalidate = subparsers.add_parser('entryvalidate', help='Remove an order if it became invalid (e.g. stop rate hit')
  parser_entryvalidate.add_argument('-id','--orderid', help='Order ID')
  parser_entryvalidate.add_argument('--demo', action='store_true', help='Use the demo account')
  parser_entryvalidate.add_argument('--real', action='store_true', help='Use the real account',default=True)
  
  #fxrmtrade
  parser_fxrmtrade = subparsers.add_parser('fxrmtrade', help='Remove a trade')
  parser_fxrmtrade.add_argument('-tid','--tradeid', help='Trade ID')
  parser_fxrmtrade.add_argument('--demo', action='store_true', help='Use the demo account')
  parser_fxrmtrade.add_argument('--real', action='store_true', help='Use the real account',default=True)
    
  #fxtr
  parser_fxtr = subparsers.add_parser('fxtr', help='Get trade details')
  parser_fxtr.add_argument('-tid','--tradeid', help='Trade ID')
  parser_fxtr.add_argument('-id','--orderid', help='Order ID')
  parser_fxtr.add_argument('--demo', action='store_true', help='Use the demo account')
  parser_fxtr.add_argument('--real', action='store_true', help='Use the real account',default=True)
  parser_fxtr.add_argument('--nosave', action='store_true', help='Dont Save the trade details')
  
  #fxmvstop
  parser_fxmvstop = subparsers.add_parser('fxmvstop', help='Move stop')
  parser_fxmvstop.add_argument('-tid','--tradeid', help='Trade ID')
  parser_fxmvstop.add_argument('-x','--stop', help='Stop')
  parser_fxmvstop.add_argument('--pips', action='store_true', help='Use pips')
  parser_fxmvstop.add_argument('--demo', action='store_true', help='Use the demo account')
  parser_fxmvstop.add_argument('--real', action='store_true', help='Use the real account',default=True)
  
  #ids
  parser_ids = subparsers.add_parser('ids', help='Refresh the IDS')
  parser_ids.add_argument('-i','--instrument', help='Instrument')
  parser_ids.add_argument('-t','--timeframe', help='Timeframe')
  parser_ids.add_argument('--full', action='store_true', help='Use the full data')
  parser_ids.add_argument('--fresh', action='store_true', help='Use the fresh data')
  
  #fxmvstopgator
  parser_fxmvstopgator = subparsers.add_parser('fxmvstopgator', help='Move stop using gator')
  parser_fxmvstopgator.add_argument('-i','--instrument', help='Instrument',default="_",required=False)
  parser_fxmvstopgator.add_argument('-t','--timeframe', help='Timeframe')
  parser_fxmvstopgator.add_argument('-tid','--tradeid', help='Trade ID')
  parser_fxmvstopgator.add_argument('--lips', action='store_true', help='Use lips')
  parser_fxmvstopgator.add_argument('--teeth', action='store_true', help='Use teeth')
  parser_fxmvstopgator.add_argument('--jaw', action='store_true', help='Use jaw')
  parser_fxmvstopgator.add_argument('--demo', action='store_true', help='Use the demo account')
  parser_fxmvstopgator.add_argument('--real', action='store_true', help='Use the real account',default=True)
  parser_fxmvstopgator.add_argument('-W','--loop_action', help='Loop action')
  
  #fxmvstopfdb
  
  parser_fxmvstopfdb = subparsers.add_parser('fxmvstopfdb', help='Move stop using fdb',epilog=fxmvstopfdb_epilog)
  parser_fxmvstopfdb.add_argument('-i','--instrument', help='Instrument',default="_",required=False)
  parser_fxmvstopfdb.add_argument('-t','--timeframe', help='Timeframe')
  parser_fxmvstopfdb.add_argument('-tid','--tradeid', help='Trade ID')
  parser_fxmvstopfdb.add_argument('--demo', action='store_true', help='Use the demo account')
  parser_fxmvstopfdb.add_argument('--real', action='store_true', help='Use the real account',default=True)
  #close the trade if the stop of fdbsignal is already hit
  parser_fxmvstopfdb.add_argument('--close', action='store_true', help='Close the trade if the stop of fdbsignal is already hit')
  parser_fxmvstopfdb.add_argument('--lips', action='store_true', help='Use lips (if no FDB signal is found, the lips will be used)')
  parser_fxmvstopfdb.add_argument('--teeth', action='store_true', help='Use teeth. If no FDB signal is found, the teeth will be used')
  parser_fxmvstopfdb.add_argument('--jaw', action='store_true', help='Use jaw (if no FDB signal is found, the jaw will be used)')
  

  parser_tidealligator = subparsers.add_parser('tide', help='Run the pto tidealligator')
  parser_tidealligator.add_argument('-i','--instrument', help='Instrument')
  parser_tidealligator.add_argument('-t','--timeframe', help='Timeframe')
  parser_tidealligator.add_argument('buysell', help='Buy or Sell')

  parser_prep_pds_01 = subparsers.add_parser('pds', help='Refresh the PDS full for an instrument and timeframe')
  parser_prep_pds_01.add_argument('-i','--instrument', help='Instrument')
  parser_prep_pds_01.add_argument('-t','--timeframe', help='Timeframe')
  parser_prep_pds_01.add_argument('--full', action='store_true', help='Use the full data')

  parser_prep_cds_05 = subparsers.add_parser('cds', help='Refresh the CDS')
  parser_prep_cds_05.add_argument('-i','--instrument', help='Instrument')
  parser_prep_cds_05.add_argument('-t','--timeframe', help='Timeframe')
  #--fresh flag to use the fresh data
  parser_prep_cds_05.add_argument('-new','--fresh', action='store_true', help='Use the fresh data')
  parser_prep_cds_05.add_argument('--full', action='store_true', help='Use the full data')
  
  parser_prep_ads_06 = subparsers.add_parser('ads', help='Refresh the ADS charts')
  parser_prep_ads_06.add_argument('-i','--instrument', help='Instrument')
  parser_prep_ads_06.add_argument('-t','--timeframe', help='Timeframe')
  #--fresh flag to use the fresh data
  parser_prep_ads_06.add_argument('-new','--fresh', action='store_true', help='Use the fresh data')
  #-pov, --save_figure_as_pov_name
  parser_prep_ads_06.add_argument('-pov','--save_figure_as_pov_name', action='store_true', help='Save the figure as pov name. Default is in ./charts/T.png')
  
  
  parser_prep_cds_06_old = subparsers.add_parser('ocds', help='Refresh the CDS from old PDS')
  parser_prep_cds_06_old.add_argument('-i','--instrument', help='Instrument')
  parser_prep_cds_06_old.add_argument('-t','--timeframe', help='Timeframe')
  parser_prep_cds_06_old.add_argument('--full', action='store_true', help='Use the full data')
  
  #tfw
  parser_tfw_cron = subparsers.add_parser('w', help='Refresh wait for timeframe')
  parser_tfw_cron.add_argument('-t','--timeframe', help='Timeframe')
  parser_tfw_cron.add_argument("-B", "--script-to-run", help="Script to run when the timeframe is reached. (.jgt/tfw.sh). ")
  parser_tfw_cron.add_argument("-X", "--exit", action="store_true", help="Exit the program when the timeframe is reached.")
  
  
  parser_prep_ttf_10 = subparsers.add_parser('ttf', help='Refresh the TTF for an instrument and timeframe')
  parser_prep_ttf_10.add_argument('-i','--instrument', help='Instrument symbol')
  parser_prep_ttf_10.add_argument('-t','--timeframe', help='Timeframe')
  parser_prep_ttf_10.add_argument('-pn','--patternname', help='Pattern Name')
  parser_prep_ttf_10.add_argument('-new','--fresh', action='store_true', help='Use the fresh data')
  parser_prep_ttf_10.add_argument('--full', action='store_true', help='Use the full data')


  
  parser_prep_mlf_22 = subparsers.add_parser('mlf', help='Refresh the MLF for an instrument and timeframe')
  parser_prep_mlf_22.add_argument('-i','--instrument', help='Instrument symbol')
  parser_prep_mlf_22.add_argument('-t','--timeframe', help='Timeframe')
  parser_prep_mlf_22.add_argument('-pn','--patternname', help='Pattern Name')
  parser_prep_mlf_22.add_argument('-tlp','--total_lagging_periods', help='Total Lagging Periods')
  parser_prep_mlf_22.add_argument('-new','--fresh', action='store_true', help='Use the fresh data')
  parser_prep_mlf_22.add_argument('--full', action='store_true', help='Use the full data')
  
  
  #ttfmxwf
  parser_post_ttfmxwf_14 = subparsers.add_parser('ttfmxwf', help='Refresh the TTF, MX and CDS for an instrument')
  parser_post_ttfmxwf_14.add_argument('-i','--instrument', help='Instrument symbol')
  parser_post_ttfmxwf_14.add_argument('-new','--fresh', action='store_true', help='Use the fresh data')

  parser_post_mx_15 = subparsers.add_parser('mx', help='Refresh the MX (using the TTF) for an instrument and timeframe')
  parser_post_mx_15.add_argument('-i','--instrument', help='Instrument symbol')
  parser_post_mx_15.add_argument('-t','--timeframe', help='Timeframe')

  parser_wf_ttf_prep_19 = subparsers.add_parser('ttfwf', help='Refresh TTF preparation for an instrument')
  parser_wf_ttf_prep_19.add_argument('-i','--instrument', help='Instrument symbol')
  parser_wf_ttf_prep_19.add_argument('-new','--fresh', action='store_true', help='Use the fresh data')
  parser_wf_ttf_prep_19.add_argument('--full', action='store_true', help='Use the full data')

  #An --autocomplete for bash
  # This argument --autocomplete is used to generate the bash autocomplete script
  ## We should not see it in the help
  parser.add_argument('--autocomplete', action='store_true',help=argparse.SUPPRESS)
  
  
  args = parser.parse_args()
  
  if args.autocomplete:
      print("fxaddorder fxrmorder entryalidate fxrmtrade fxtr fxmvstop ids fxmvstopgator fxmvstopfdb")#Generating bash autocomplete script...")
      
      exit()
      
  #if no arguments are passed, print help
  if not vars(args).get('command'):
    parser.print_help()
    parser.exit()
  
  
  
  if args.command == 'tide':
    tide(args.instrument, args.timeframe, args.buysell)
  elif args.command == 'fxaddorder':
    fxaddorder(args.instrument, args.lots, args.rate, args.buysell, args.stop, args.demo,args.pips)
  elif args.command == 'fxrmorder':
    fxrmorder(args.orderid, args.demo)
  elif args.command == 'entryvalidate':
    entryvalidate(args.orderid, args.demo)
  elif args.command == 'fxrmtrade':
    fxrmtrade(args.tradeid, args.demo)
  elif args.command == 'fxtr':
    fxtr(args.tradeid,args.orderid,args.demo,not args.nosave)
  elif args.command == 'fxmvstop':
    fxmvstop(args.tradeid, args.stop, args.pips, args.demo)
  elif args.command == 'ids':
    ids(args.instrument, args.timeframe,args.full,args.fresh)
  elif args.command == 'fxmvstopgator':
    lips_value = True if not args.lips and not args.teeth and not args.jaw else False
    fxmvstopgator(args.instrument, args.timeframe, args.tradeid, lips_value,args.teeth,args.jaw,args.demo,args.loop_action)
  elif args.command == 'fxmvstopfdb':
    fxmvstopfdb(args.instrument, args.timeframe, args.tradeid, args.demo,args.close,args.lips,args.teeth,args.jaw)
  elif args.command == 'w':
    w(args.timeframe,args.script_to_run,args.exit)
  elif args.command == 'pds':
    pds(args.instrument, args.timeframe,)
  elif args.command == 'cds':
    cds(args.instrument, args.timeframe, args.fresh,args.full)
  elif args.command == 'ads':
    use_pov_name = args.save_figure_as_pov_name if args.save_figure_as_pov_name else False
    tc_flag = False if use_pov_name else True
    
    ads(args.instrument, args.timeframe, args.fresh,pov=use_pov_name,tc=tc_flag)
  elif args.command == 'ocds':
    ocds(args.instrument, args.timeframe)
  elif args.command == 'ttf':
    ttf(args.instrument, args.timeframe,args.patternname,args.columns_list_from_higher_tf,args.fresh,args.full)
  elif args.command == 'mlf':
    mlf(args.instrument, args.timeframe,args.patternname,args.total_lagging_periods,args.fresh,args.full)
  elif args.command == 'ttfmxwf':
    ttfmxwf(args.instrument, args.fresh)
  elif args.command == 'mx':
    mx(args.instrument, args.timeframe)
  elif args.command == 'ttfwf':
    ttfwf(args.instrument, args.fresh,args.full)

if __name__ == "__main__":
  main()




"""
Here we would have common data functions such as : write_patternname_columns_list and read_patternname_columns_list

Therefore: 
 * ttfcli would  
"""
import os
import pandas as pd
from mlutils import get_basedir,get_outfile_fullpath,get_list_of_files_in_ns
from mlconstants import PATTERN_NS, TTF_NOT_NEEDED_COLUMNS_LIST, default_columns_to_get_from_higher_tf,TTF_DTYPE_DEFINITION



settings:dict = {}

def load_settings(custom_path=None,args=None):
  global settings
  if args is not None:
    if hasattr(args,"jgtcommon_settings"):
      _settings=getattr(args,"jgtcommon_settings")
      settings.update(_settings)
  if settings is None or len(settings)==0:
    from jgtutils.jgtcommon import load_settings as jgtcommon_load_settings
    settings = jgtcommon_load_settings(custom_path)
  return settings

def get_settings():
  global settings
  if settings is None or len(settings)==0:
    settings=load_settings()
  return settings

if __name__ != "__main__": 
  load_settings()


#ttf


def get_ttf_outfile_fullpath(i,t,use_full=True,suffix="",ns="ttf",pn="ttf"):
  return get_outfile_fullpath(i,t,use_full,ns,pn=pn,suffix=suffix)


def write_patternname_columns_list(i,t,use_full=True,columns_list_from_higher_tf=None,pn="ttf",ns="ttf",suffix="_columns"):
  if columns_list_from_higher_tf is None:
    columns_list_from_higher_tf = default_columns_to_get_from_higher_tf
  output_filename=get_ttf_outfile_fullpath(i,t,use_full,suffix=suffix,pn=pn,ns=ns)
  with open(output_filename, 'w') as f:
    for item in columns_list_from_higher_tf:
      f.write("%s\n" % item)
  print(f"    Pattern:{pn} Output columns :'{output_filename}'")
  return output_filename

def read_patternname_columns_list(i,t,use_full=True,pn="ttf",ns="ttf",suffix="_columns")->list:
  output_filename=get_ttf_outfile_fullpath(i,t,use_full,suffix=suffix,pn=pn,ns=ns)
  #if not os.path.exists(output_filename), we would try to read from the pndata...
  flag_we_have_pattern_stored_specific_to_pov = os.path.exists(output_filename)
  #flag_we_have_pattern_stored_specific_to_pov=False # BYPASS FOR DEVELOPMENT
  if not flag_we_have_pattern_stored_specific_to_pov:
    raw_columns = pndata__read_new_pattern_columns_list(pn=pn)
    raise NotImplementedError("We Require to Add the Timeframe Columns to the Pattern")
    return raw_columns
  with open(output_filename, 'r') as f:
    columns_list_from_higher_tf = f.readlines()
  columns_list_from_higher_tf = [x.strip() for x in columns_list_from_higher_tf]
  return columns_list_from_higher_tf

def create_filebase_from_patternname(i,t,pn="ttf")->str:
  ifn=i.replace("/","-")
  output_filename = f"{ifn}_{t}_{pn}"
  return output_filename.replace("__","_").replace('"','')

def create_filensbase_from_patternname(i,t,pn="ttf",ns="ttf")->str:
  filebase=create_filebase_from_patternname(i,t,pn)
  return f"{ns}/{filebase}"


#@STCIssue Future Generic we would use for other patterns (ex.  targets/mx)
def read_pattern_raw(i, t,pn,ns, use_full=True)->pd.DataFrame:

  outfile_fullpath=get_outfile_fullpath(i,t,use_full,ns,pn=pn)
  df=pd.read_csv(outfile_fullpath, index_col=0,dtype=TTF_DTYPE_DEFINITION)
  return df


def read_ttf_pattern_raw(i, t, use_full=True,pn="ttf",ns="ttf")->pd.DataFrame:
  ttf_outfile_fullpath=get_ttf_outfile_fullpath(i,t,pn=pn,ns=ns,use_full=use_full)
  df=pd.read_csv(ttf_outfile_fullpath, index_col=0,dtype=TTF_DTYPE_DEFINITION)
  return df
  

def read_ttf_feature_columns_only_from_pattern(i, t, use_full=True,pn="ttf",ns="ttf"):
  df=read_ttf_pattern_raw(i, t, use_full=use_full,pn=pn,ns=ns)
  pattern_columns_list:list=read_patternname_columns_list(i,t,pn=pn,ns=ns)
  #keep only the columns from the list
  df=df[pattern_columns_list]
  return df


MLF_NS = "mlf"

def write_mlf_pattern_lagging_columns_list(i, t, use_full=True, pn="ttf", lagging_columns=None):
  print("WARN::write_mlf_pattern_lagging_columns_list::WILL BE DEPRECATED")
  write_patternname_columns_list(i,t,use_full,lagging_columns,pn=pn,ns=MLF_NS)

def read_mlf_pattern_lagging_columns_list(i, t, use_full=True, pn="ttf"):
  #@STCIssue UPGRADE TO USE THE PATTERNNAME / Settings - REPLACE by : pndata__read_new_pattern_columns_list_with_htf_and_lags_using_settings
  #lagging_columns=read_patternname_columns_list(i,t,use_full,pn=pn,ns=MLF_NS)
  lagging_columns=pndata__read_new_pattern_columns_list_with_htf_and_lags_using_settings(t,pn,lag_period=1, total_lagging_periods=5, out_lag_midfix_str='_lag_')
  return lagging_columns

def read_mlf_pattern_raw(i, t, use_full=True,pn="ttf"):
  df=read_ttf_pattern_raw(i, t, use_full=use_full,pn=pn,ns=MLF_NS)
  return df

def pto_read_pattern_columns_list_with_higher_tf(i, t, use_full=True, pn="ttf"):
  raise NotImplementedError("""
  #@STCGoal See : pndata__read_new_pattern_columns_list_with_htf
                            """)
#   columns_list_from_higher_tf = read_patternname_columns_list(i,t,use_full,pn=pn,ns="ttf")
#   return columns_list_from_higher_tf


def read_mlf_for_pattern(i, t, use_full=True,pn="ttf"):
  df=read_ttf_pattern_raw(i, t, use_full=use_full,pn=pn,ns=MLF_NS)
  return df

def read_mlf_feature_columns_only_from_pattern(i, t, use_full=True,pn="ttf"):
  df=read_mlf_for_pattern(i, t, use_full=use_full,pn=pn)
  lagging_columns_list:list=read_mlf_pattern_lagging_columns_list(i,t,pn=pn,use_full=use_full)
  #keep only the columns from the list
  df=df[lagging_columns_list]
  return df



def pndata__write_new_pattern_columns_list(columns_list_from_higher_tf, pn, suffix=""):
    if columns_list_from_higher_tf is None:
        raise ValueError("columns_list_from_higher_tf cannot be None")
    pattern_filename=get_outfile_fullpath("-","-",True,PATTERN_NS,pn=pn,suffix=suffix)
    with open(pattern_filename, 'w') as f:
        for item in columns_list_from_higher_tf:
            f.write("%s\n" % item)
    print(f"Pattern: {pn} Output columns: '{pattern_filename}'")
    return pattern_filename

def pndata__get_all_patterns_from_settings(args=None):
  global settings
  if settings is None or len(settings)==0:
    settings=load_settings(args=args)
  if "patterns" in settings:
    patterns_data = settings["patterns"]
    return patterns_data
  return {}

#get_list_of_files_in_ns
#@DeprecationWarning("Use of JGT Settings to Define patterns")
def pndata__get_all_patterns(use_full=True,output_type="object",args=None):
  list_of_files=get_list_of_files_in_ns(use_full,PATTERN_NS)
  #read all files in the directory and make a dictionary of the pattern names and their columns
  o={}
  for f in list_of_files:
    pn=f.split(".")[0]
    #print(f"Pattern: {pn}")
    columns_of_the_pattern=pndata__read_new_pattern_columns_list(pn=pn,args=args)
    #print(f"Columns: {columns_of_the_pattern}")
    p={"columns":columns_of_the_pattern}
    o[pn]=p
  
  patterns_in_settings=pndata__get_all_patterns_from_settings(args=args)
  o.update(patterns_in_settings)
  
  if output_type=="json":
    import json
    return json.dumps(o, indent=4)
  if output_type=="md":
    return _patterns_dictionary_to_markdown(o)
  return o

def _patterns_dictionary_to_markdown(patterns_dictionary):
    _out_string = "| Pattern | Columns |"
    _out_string += "\n"+ "| --- | --- |"
                
    for k,v in patterns_dictionary.items():
        _line = f"| {k} | "
        _columns_list = v['columns']
        _len_of_columns = len(_columns_list)
        for c in _columns_list:
            coma_if_not_last = "," if _columns_list.index(c) < _len_of_columns-1 else ""
            _line += f"{c}{coma_if_not_last}"
        _line += " |"
        #{v['columns']} |"
        _out_string += "\n"+_line
    return _out_string 


def pndata__read_new_pattern_columns_list(pn, suffix="",args=None)->list[str]:
    global settings
    if settings is None or len(settings)==0:
        settings=load_settings(args=args)

    
    pattern_filename=get_outfile_fullpath("-","-",True,PATTERN_NS,pn=pn,suffix=suffix)
    #Support reading from the args if not none
    if args is not None:
      if hasattr(args,"columns_list_from_higher_tf") and args.columns_list_from_higher_tf is not None:
        return args.columns_list_from_higher_tf
    
    if "patterns" in settings and pn in settings["patterns"]:
      patterns=settings["patterns"]
      if pn in patterns:
        pattern_columns_values = patterns[pn]["columns"]
        return pattern_columns_values
    
    if not os.path.exists(pattern_filename):
      raise FileNotFoundError(f"File {pattern_filename} does not exist.  Use -clh <col1 col2 ...> to create it.")
    
    with open(pattern_filename, 'r') as f:
        columns_list = [line.strip() for line in f]
    #print(f"Pattern: {pn} Read columns: {columns_list}")
    return columns_list

from jgtutils import jgtpov as jpov
def _ptottf__make_htf_created_columns_array(workset,t,columns_list_from_higher_tf):

  created_columns=[]
  for c in columns_list_from_higher_tf:
    for k in workset:
      if not c in created_columns: 
        created_columns.append(c)
      new_col_name = c+"_"+k
      if k != t:
        if not new_col_name in created_columns: 
          created_columns.append(new_col_name)
  return created_columns

def pndata__read_new_pattern_columns_list_with_htf(t:str,pn:str, suffix="",args=None):
  #Pattern and its columns without the Higher Timeframe features
  raw_pattern_column_list:list[str]=pndata__read_new_pattern_columns_list(pn=pn,suffix=suffix,args=args)
  povs = jpov.get_higher_tf_array(t)
  created_columns=_ptottf__make_htf_created_columns_array(povs,t,raw_pattern_column_list)
  return created_columns

def pndata__read_new_pattern_columns_list_with_htf_and_lags_using_settings(t,pn, lag_period=1, total_lagging_periods=5, out_lag_midfix_str='_lag_',args=None):
  columns_to_add_lags_to=pndata__read_new_pattern_columns_list_with_htf(t,pn,args=args)
  
  new_cols = _get_lagging_columns_list(columns_to_add_lags_to, lag_period=lag_period, total_lagging_periods=total_lagging_periods, out_lag_midfix_str=out_lag_midfix_str)
  
  extended_columns_list = columns_to_add_lags_to+new_cols
  return extended_columns_list


def _get_lagging_columns_list(columns_to_add_lags_to, lag_period=1, total_lagging_periods=5, out_lag_midfix_str='_lag_'):
    new_cols = []  # List to hold new lagging columns
    for col in columns_to_add_lags_to:
        for j in range(1, total_lagging_periods + 1):
            lag_col_name = _create_lag_column_name(out_lag_midfix_str, col, j)
            new_cols.append(lag_col_name)
    return new_cols
  
def _create_lag_column_name(out_lag_midfix_str, col, j):
    return f'{col}{out_lag_midfix_str}{j}'


class JGTML240615Config:
    """
    Configuration class for the JGTML240615 prototype.
    """

    def __init__(self,
                 instrument: str = "GBP/USD",
                 timeframe: str = "H4",
                 buysell: str = "S",
                 regenerate_cds: bool = False,
                 result_drop_base_override: str = None,
                 source_dataset_archival_path_override: str = None,
                 quiet: bool = False,
                 print_abstract: bool = False,
                 use_ttf_rather_than_cds: bool = False,
                 result_file_basename: str = "jgtml_obsds_240515_SIGNALS.result",
                 data_dir_override: str = None,
                 use_fresh: bool = True,
                 mfi_flag: bool = True,
                 balligator_flag: bool = True,
                 talligator_flag: bool = True,
                 ):
        """
        Initialize the JGTML240615Config object.

        Args:
            instrument: The financial instrument to analyze.
            timeframe: The timeframe of the data to analyze.
            buysell: The direction of the signals to analyze ("S" for sell, "B" for buy).
            regenerate_cds: Whether to regenerate the CDS data.
            result_drop_base_override: The path to the directory where the results will be saved.
            source_dataset_archival_path_override: The path to the directory where the source dataset is archived.
            quiet: Whether to suppress output.
            print_abstract: Whether to print the abstract.
            use_ttf_rather_than_cds: Whether to use TTF data as the source of data.
            result_file_basename: The base name of the result files.
            data_dir_override: The path to the directory containing the data.
            use_fresh: Whether to use fresh data.
            mfi_flag: Whether to include MFI data.
            balligator_flag: Whether to include Big Alligator data.
            talligator_flag: Whether to include Tide Alligator data.
        """

        self.instrument = instrument
        self.timeframe = timeframe
        self.buysell = buysell
        self.regenerate_cds = regenerate_cds
        self.result_drop_base_override = result_drop_base_override
        self.source_dataset_archival_path_override = source_dataset_archival_path_override
        self.quiet = quiet
        self.print_abstract = print_abstract
        self.use_ttf_rather_than_cds = use_ttf_rather_than_cds
        self.result_file_basename = result_file_basename
        self.data_dir_override = data_dir_override
        self.use_fresh = use_fresh
        self.mfi_flag = mfi_flag
        self.balligator_flag = balligator_flag
        self.talligator_flag = talligator_flag


import os
import sys

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))

import pandas as pd
pd.options.mode.copy_on_write = True

import anhelper

from jgtutils.jgtconstants import ZONE_INT,ZONE_BUY_ID,ZONE_SELL_ID,ZONE_NEUTRAL_ID,ZONE_BUY_STR,ZONE_SELL_STR,ZONE_NEUTRAL_STR
from jgtutils.colconverthelper import zone_str_to_id,get_zone_features_column_list_by_timeframe

from mlconstants import ZONE_DEFAULT_COLNAME
  
def column_zone_str_in_dataframe_to_id(df:pd.DataFrame,t:str,inplace=False,zone_colname=""):
    """
    Convert the ZONE columns from str to id in the dataframe.
    
    Parameters:
    df (pd.DataFrame): The dataframe to convert the ZONE columns from str to id.
    t (str): The timeframe to convert the ZONE columns from str to id.
    inplace (bool): If True, the conversion is done in place. If False, a copy of the dataframe is returned with the conversion done.
    zone_colname (str): The name of the ZONE column to use. If not provided, the default ZCOL is used. (Planning to use ZONE_SIGNAL)
    
    Returns:
    pd.DataFrame: The dataframe with the ZONE columns converted from str to id.
    
    """
    if zone_colname=="":
        zone_colname=ZONE_DEFAULT_COLNAME
        
    if not inplace:
        df = df.copy()
    zcol_features_columns_list = get_zone_features_column_list_by_timeframe(t,zone_colname)
    for col_name in zcol_features_columns_list:
        df[col_name] = df[col_name].apply(lambda x: int(zone_str_to_id(x)))
          #zonecolor_str_to_id)
    return df

def _zoneint_add_lagging_feature(df: pd.DataFrame, t, lag_period=1, total_lagging_periods=5,out_lag_midfix_str='_lag_',inplace=True,zone_colname=""):
    if zone_colname=="":
        zone_colname=ZONE_DEFAULT_COLNAME
        
    if not inplace:
        df = df.copy()
    columns_to_add_lags_to = get_zone_features_column_list_by_timeframe(t,zone_colname)
    #columns_to_add_lags_to.append(zone_colname) #We want a lag for the current TF
    anhelper.add_lagging_columns(df, columns_to_add_lags_to, lag_period, total_lagging_periods, out_lag_midfix_str)
    for col in columns_to_add_lags_to:#@STCIssue Isn't that done already ???  Or it thinks they are Double !!!!
        for j in range(1, total_lagging_periods + 1):
            df[f'{col}{out_lag_midfix_str}{j}']=df[f'{col}{out_lag_midfix_str}{j}'].astype(int)
    return df
    

def wf_mk_zone_ready_dataset__240708(df: pd.DataFrame, t, lag_period=1, total_lagging_periods=5,out_lag_midfix_str='_lag_',inplace=True,zone_colname=""):
    if zone_colname=="":
        zone_colname=ZONE_DEFAULT_COLNAME
        
    if not inplace:
        df = df.copy()
    #column_zone_str_in_dataframe_to_id(df,t,inplace=True,zone_colname=zone_colname)
    _zoneint_add_lagging_feature(df,t,lag_period=lag_period, total_lagging_periods=total_lagging_periods,out_lag_midfix_str=out_lag_midfix_str,inplace=True,zone_colname=zone_colname)
    return df




import pandas as pd
# Use jgtconstants column names from jgtutils
from jgtutils.jgtconstants import LOW,HIGH,FDBB,FDBS,BJAW,BLIPS,BTEETH,JAW,TEETH,LIPS,FDB_TARGET,VECTOR_AO_FDBS_COUNT,VECTOR_AO_FDBB_COUNT,VECTOR_AO_FDB_COUNT,TJAW,TLIPS,TTEETH


def get_alligator_column_names_from_ctx_name(ctx_name):
    if ctx_name=="ripple" or ctx_name=="normal":
        cteeth_colname = TEETH
        clips_colname = LIPS
        cjaw_colname = JAW

    if ctx_name=="tide":
        cteeth_colname = TTEETH
        clips_colname = TLIPS
        cjaw_colname = TJAW


    if ctx_name=="big":
        cteeth_colname = BTEETH
        clips_colname = BLIPS
        cjaw_colname = BJAW
    return cteeth_colname,clips_colname,cjaw_colname





def generate_column_names_for_direction(ctx_name,direction)->list[str]:
  
    out_sig_is_in_ctx_teeth_sum_colname = f"{direction}_sig_is_in_{ctx_name}_teeth_sum"
    out_sig_ctx_mouth_is_open_and_signal_is_in_ctx_lips_sum_colname = f"{direction}_sig_{ctx_name}_mouth_is_open_and_in_{ctx_name}_lips"
    out_sig_ctx_mouth_is_open_and_signal_is_in_ctx_teeth_sum_colname = f"{direction}_sig_{ctx_name}_mouth_is_open_and_in_{ctx_name}_teeth_sum"
    out_sig_normal_mouth_is_open_sum_colname = "sig_normal_mouth_is_open"

    
    return {
        "sig_is_in_ctx_teeth_sum": out_sig_is_in_ctx_teeth_sum_colname,
        "sig_ctx_mouth_is_open_and_in_ctx_lips": out_sig_ctx_mouth_is_open_and_signal_is_in_ctx_lips_sum_colname,
        "sig_ctx_mouth_is_open_and_in_ctx_teeth_sum": out_sig_ctx_mouth_is_open_and_signal_is_in_ctx_teeth_sum_colname,
        "sig_normal_mouth_is_open_sum": out_sig_normal_mouth_is_open_sum_colname
    }



# Base filtering
def _filter_sig_is_out_of_normal_mouth_sell(dfsrc):
    df_sig_is_out_of_normal_mouth = dfsrc[dfsrc[LOW] > dfsrc[LIPS]].copy()
    df_sig_is_out_of_normal_mouth = df_sig_is_out_of_normal_mouth[df_sig_is_out_of_normal_mouth[LOW] > df_sig_is_out_of_normal_mouth[TEETH]]
    return df_sig_is_out_of_normal_mouth

def _filter_sig_normal_mouth_is_open_sell(dfsrc):
    df_sig_normal_mouth_is_open = dfsrc[dfsrc[JAW] < dfsrc[TEETH]].copy()
    df_sig_normal_mouth_is_open = df_sig_normal_mouth_is_open[df_sig_normal_mouth_is_open[TEETH] < df_sig_normal_mouth_is_open[LIPS]]
    df_sig_normal_mouth_is_open = df_sig_normal_mouth_is_open[df_sig_normal_mouth_is_open[JAW] < df_sig_normal_mouth_is_open[LIPS]]
    return df_sig_normal_mouth_is_open

def _filter_sig_is_out_of_normal_mouth_buy(dfsrc):
    df_sig_is_out_of_normal_mouth = dfsrc[dfsrc[HIGH] < dfsrc[LIPS]].copy()
    df_sig_is_out_of_normal_mouth = df_sig_is_out_of_normal_mouth[df_sig_is_out_of_normal_mouth[HIGH] < df_sig_is_out_of_normal_mouth[TEETH]]
    return df_sig_is_out_of_normal_mouth

def _filter_sig_normal_mouth_is_open_buy(dfsrc):
    df_sig_normal_mouth_is_open = dfsrc[dfsrc[JAW] > dfsrc[TEETH]].copy()
    df_sig_normal_mouth_is_open = df_sig_normal_mouth_is_open[df_sig_normal_mouth_is_open[TEETH] > df_sig_normal_mouth_is_open[LIPS]]
    df_sig_normal_mouth_is_open = df_sig_normal_mouth_is_open[df_sig_normal_mouth_is_open[JAW] > df_sig_normal_mouth_is_open[LIPS]]
    return df_sig_normal_mouth_is_open



def filter_sig_is_out_of_normal_mouth(dfsrc,bs,ctx_name):
  teval_colname, cteeth_colname = __get_column_names_from_context(bs, ctx_name)
  raise NotImplementedError("Not implemented yet")
  


def filter_sig_is_in_ctx_teeth(dfsrc,bs,ctx_name):
  teval_colname,cteeth_colname,clips_colname,cjaw_colname = __get_column_names_from_context(bs, ctx_name)
  df_sig_is_in_ctx_teeth = _filter_sig_is_in_ctx_teeth_sell(dfsrc, cteeth_colname, teval_colname) if bs=="S" else _filter_sig_is_in_ctx_teeth_buy(dfsrc, cteeth_colname, teval_colname)
  return df_sig_is_in_ctx_teeth


def filter_sig_ctx_mouth_is_open_and_in_ctx_teeth(dfsrc,bs,ctx_name):
  teval_colname,cteeth_colname,clips_colname,cjaw_colname = __get_column_names_from_context(bs, ctx_name)
  df_sig_ctx_mouth_is_open_and_in_ctx_teeth=_filter_sig_ctx_mouth_is_open_and_in_ctx_teeth_sell(dfsrc, cteeth_colname, clips_colname, cjaw_colname, teval_colname) if bs=="S" else _filter_sig_ctx_mouth_is_open_and_in_ctx_teeth_buy(dfsrc, cteeth_colname, clips_colname, cjaw_colname, teval_colname)    
  return df_sig_ctx_mouth_is_open_and_in_ctx_teeth

def filter_sig_ctx_mouth_is_open_and_in_ctx_lips(dfsrc,bs,ctx_name):
  teval_colname,cteeth_colname,clips_colname,cjaw_colname = __get_column_names_from_context(bs, ctx_name)
  #print("teval_colname",teval_colname)
  #print("cteeth_colname",cteeth_colname)
  #print("clips_colname",clips_colname)
  #print("cjaw_colname",cjaw_colname)
  df_sig_ctx_mouth_is_open_and_in_ctx_lips=_filter_sig_ctx_mouth_is_open_and_in_ctx_lips_sell(dfsrc, cteeth_colname, clips_colname, cjaw_colname, teval_colname)  if bs=="S" else _filter_sig_ctx_mouth_is_open_and_in_ctx_lips_buy(dfsrc, cteeth_colname, clips_colname, cjaw_colname, teval_colname)
  return df_sig_ctx_mouth_is_open_and_in_ctx_lips


def __new_ctx_colname(ctx_name,bs,ctx_basename,tolower=False):
  new_col = ctx_basename.replace("_ctx_","_"+ctx_name+"_")  + "_"+bs#.lower()
  fixed_return = new_col.replace("__","_")
  
  return fixed_return if not tolower else fixed_return.lower()

def add_sig_ctx_mouth_is_open_and_in_ctx_lips(dfsrc, bs,ctx_name, new_col=None,ctx_basename = "sig_ctx_mouth_is_open_and_in_ctx_lips")->pd.DataFrame:
  teval_colname,cteeth_colname,clips_colname,cjaw_colname = __get_column_names_from_context(bs, ctx_name)
  if new_col is None:    
    new_col = __new_ctx_colname(ctx_name,bs,ctx_basename)
    
  # Create a boolean mask based on the conditions
  if bs=="S":
    mask = (
        (dfsrc[teval_colname] < dfsrc[clips_colname]) &
        (dfsrc[clips_colname] < dfsrc[cteeth_colname]) &
        (dfsrc[cteeth_colname] < dfsrc[cjaw_colname])
    )
  else:
    mask = (
      (dfsrc[teval_colname] > dfsrc[clips_colname]) &
      (dfsrc[clips_colname] > dfsrc[cteeth_colname]) &
      (dfsrc[cteeth_colname] > dfsrc[cjaw_colname])
      )
    
  
  # Add a new column based on the mask
  dfsrc[new_col] = mask.astype(int)
  
  return dfsrc


# CTX filtering

def _filter_sig_is_in_ctx_teeth_sell(dfsrc, cteeth_colname, teval_colname):
    df_sig_is_in_ctx_teeth=dfsrc[
    dfsrc[teval_colname] > dfsrc[cteeth_colname]
    ].copy()
    return df_sig_is_in_ctx_teeth

def _filter_sig_is_in_ctx_teeth_buy(df_sig_is_out_of_normal_mouth, cteeth_colname, teval_colname):
    df_sig_is_in_ctx_teeth=df_sig_is_out_of_normal_mouth[
    df_sig_is_out_of_normal_mouth[teval_colname] < df_sig_is_out_of_normal_mouth[cteeth_colname]
    ].copy()
    return df_sig_is_in_ctx_teeth

  

"""
df_sig_is_in_ctx_teeth = filter_sig_is_in_ctx_teeth_sell(df_sig_is_out_of_normal_mouth, cteeth_colname, teval_colname) if bs=="S" else filter_sig_is_in_ctx_teeth_buy(df_sig_is_out_of_normal_mouth, cteeth_colname, teval_colname)

df_sig_ctx_mouth_is_open_and_in_ctx_teeth=filter_sig_ctx_mouth_is_open_and_in_ctx_teeth_sell(df_sig_is_out_of_normal_mouth, cteeth_colname, clips_colname, cjaw_colname, teval_colname) if bs=="S" else filter_sig_ctx_mouth_is_open_and_in_ctx_teeth_buy(df_sig_is_out_of_normal_mouth, cteeth_colname, clips_colname, cjaw_colname, teval_colname)    

df_sig_ctx_mouth_is_open_and_in_ctx_lips=filter_sig_ctx_mouth_is_open_and_in_ctx_lips_sell(df_sig_is_out_of_normal_mouth, cteeth_colname, clips_colname, cjaw_colname, teval_colname)  if bs=="S" else filter_sig_ctx_mouth_is_open_and_in_ctx_lips_buy(df_sig_is_out_of_normal_mouth, cteeth_colname, clips_colname, cjaw_colname, teval_colname)
"""

def _filter_sig_ctx_mouth_is_open_and_in_ctx_teeth_sell(dfsrc, cteeth_colname, clips_colname, cjaw_colname, teval_colname):
    df_sig_ctx_mouth_is_open_and_in_ctx_teeth = dfsrc[
        dfsrc[teval_colname] > dfsrc[cteeth_colname]
        ].copy()

    df_sig_ctx_mouth_is_open_and_in_ctx_teeth = df_sig_ctx_mouth_is_open_and_in_ctx_teeth[  
        df_sig_ctx_mouth_is_open_and_in_ctx_teeth[clips_colname] < df_sig_ctx_mouth_is_open_and_in_ctx_teeth[cteeth_colname]
        ]
    df_sig_ctx_mouth_is_open_and_in_ctx_teeth = df_sig_ctx_mouth_is_open_and_in_ctx_teeth[  
        df_sig_ctx_mouth_is_open_and_in_ctx_teeth[cteeth_colname] < df_sig_ctx_mouth_is_open_and_in_ctx_teeth[cjaw_colname]
        ]
    return df_sig_ctx_mouth_is_open_and_in_ctx_teeth

def _filter_sig_ctx_mouth_is_open_and_in_ctx_teeth_buy(dfsrc, cteeth_colname, clips_colname, cjaw_colname, teval_colname):
    df_sig_ctx_mouth_is_open_and_in_ctx_teeth = dfsrc[
        dfsrc[teval_colname] < dfsrc[cteeth_colname]
        ].copy()

    df_sig_ctx_mouth_is_open_and_in_ctx_teeth = df_sig_ctx_mouth_is_open_and_in_ctx_teeth[  
        df_sig_ctx_mouth_is_open_and_in_ctx_teeth[clips_colname] > df_sig_ctx_mouth_is_open_and_in_ctx_teeth[cteeth_colname]
        ]
    df_sig_ctx_mouth_is_open_and_in_ctx_teeth = df_sig_ctx_mouth_is_open_and_in_ctx_teeth[  
        df_sig_ctx_mouth_is_open_and_in_ctx_teeth[cteeth_colname] > df_sig_ctx_mouth_is_open_and_in_ctx_teeth[cjaw_colname]
        ]
        
    return df_sig_ctx_mouth_is_open_and_in_ctx_teeth


def _filter_sig_ctx_mouth_is_open_and_in_ctx_lips_sell(dfsrc, cteeth_colname, clips_colname, cjaw_colname, teval_colname):
    df_sig_ctx_mouth_is_open_and_in_ctx_lips = dfsrc[
        dfsrc[teval_colname] < dfsrc[clips_colname]
        ].copy()

    # the CMouth is Openned
    df_sig_ctx_mouth_is_open_and_in_ctx_lips = df_sig_ctx_mouth_is_open_and_in_ctx_lips[
        df_sig_ctx_mouth_is_open_and_in_ctx_lips[clips_colname] < df_sig_ctx_mouth_is_open_and_in_ctx_lips[cteeth_colname]
        ]
    df_sig_ctx_mouth_is_open_and_in_ctx_lips = df_sig_ctx_mouth_is_open_and_in_ctx_lips[
        df_sig_ctx_mouth_is_open_and_in_ctx_lips[cteeth_colname] < df_sig_ctx_mouth_is_open_and_in_ctx_lips[cjaw_colname]
        ]
    return df_sig_ctx_mouth_is_open_and_in_ctx_lips



def _filter_sig_ctx_mouth_is_open_and_in_ctx_lips_buy(df_sig_is_out_of_normal_mouth, cteeth_colname, clips_colname, cjaw_colname, teval_colname):
    df_sig_ctx_mouth_is_open_and_in_ctx_lips = df_sig_is_out_of_normal_mouth[
        df_sig_is_out_of_normal_mouth[teval_colname] > df_sig_is_out_of_normal_mouth[clips_colname]
        ].copy()

    # the BMouth is Openned
    df_sig_ctx_mouth_is_open_and_in_ctx_lips = df_sig_ctx_mouth_is_open_and_in_ctx_lips[
        df_sig_ctx_mouth_is_open_and_in_ctx_lips[clips_colname] > df_sig_ctx_mouth_is_open_and_in_ctx_lips[cteeth_colname]
        ]
    df_sig_ctx_mouth_is_open_and_in_ctx_lips = df_sig_ctx_mouth_is_open_and_in_ctx_lips[
        df_sig_ctx_mouth_is_open_and_in_ctx_lips[cteeth_colname] > df_sig_ctx_mouth_is_open_and_in_ctx_lips[cjaw_colname]
        ]
        
    return df_sig_ctx_mouth_is_open_and_in_ctx_lips



def __get_column_names_from_context(bs, ctx_name):
    teval_colname = LOW if bs=="S" else HIGH
    cteeth_colname,clips_colname,cjaw_colname=get_alligator_column_names_from_ctx_name(ctx_name)
    return teval_colname,cteeth_colname,clips_colname,cjaw_colname


"""
@STCGoal Wrap what MX does as Services

- Create MX
- Read MX (and create if doesn't exist and if we have a corresponding TTF PatternsData in the DB, we can also create the TTF and MLF)
- Update MX
- Create TTF/MLF Patterns (data:  PatternName, PatternsColumns) 

Dependencies:
- MXRequest
- MLF

Target Usage:
- mxcli
- mxapi
"""



import pandas as pd

pd.options.mode.copy_on_write = True
#PerformanceWarning
import warnings
warnings.filterwarnings("ignore", category=FutureWarning)
warnings.filterwarnings("ignore", category=pd.errors.PerformanceWarning)



from mlconstants import CONVERTION_EXCLUDED_COLUMNS


def add_lagging_columns(dfsrc: pd.DataFrame, columns_to_add_lags_to, lag_period=1, total_lagging_periods=5, out_lag_midfix_str='_lag_', dropna=True)->pd.DataFrame:
    new_cols = []  # List to hold new lagging columns
    convertion_included=[] # Those columns that are not in the excluded list (their parent somehow, we want to convert them to int when we add lagging columns but not those that are in the excluded list (mostly because they are double))
    for col in columns_to_add_lags_to:
        if col not in dfsrc.columns:
            print("WARN::Column not in dataframe:", col, " Skipping")
            continue
        else:
            for j in range(1, total_lagging_periods + 1):
                # Create new lagging column without modifying the original DataFrame
                lag_col_name = _create_lag_column_name(out_lag_midfix_str, col, j)
                lag_col = dfsrc[col].shift(j * lag_period).rename(lag_col_name)
                new_cols.append(lag_col)
                if col not in CONVERTION_EXCLUDED_COLUMNS:
                    convertion_included=convertion_included+[lag_col_name]

    # Concatenate the original DataFrame with the new columns
    dfsrc = pd.concat([dfsrc] + new_cols, axis=1)

    # Drop rows where there are no lag values
    if dropna:
        dfsrc.dropna(inplace=True)

    # Convert lagging columns to integer
    for col in columns_to_add_lags_to:
        for j in range(1, total_lagging_periods + 1):
            lag_col_name = _create_lag_column_name(out_lag_midfix_str, col, j)
            if lag_col_name in convertion_included: #Hopefully, double columns wont be converted
              dfsrc[lag_col_name] = dfsrc[lag_col_name].astype(int)
            else:
              pass

    return dfsrc

from mldatahelper import _create_lag_column_name

  
def add_lagging_columns_V1(dfsrc: pd.DataFrame, columns_to_add_lags_to, lag_period=1, total_lagging_periods=5, out_lag_midfix_str='_lag_',dropna=True):
  for col in columns_to_add_lags_to:
    if col not in dfsrc.columns:
      print("WARN::Column not in dataframe:", col, " Skipping")
      pass
    else:
      for j in range(1, total_lagging_periods + 1):
      # Using .loc to ensure the operation is done without warnings
        dfsrc.loc[:, f'{col}{out_lag_midfix_str}{j}'] = dfsrc[col].shift(j * lag_period)
  # Drop rows where there are no lag values
  if dropna:
    dfsrc.dropna(inplace=True)
    
  for col in columns_to_add_lags_to: #@STCIssue Isn't that done already ???  Or it thinks they are Double !!!!
    for j in range(1, total_lagging_periods + 1):
      dfsrc[f'{col}{out_lag_midfix_str}{j}']=dfsrc[f'{col}{out_lag_midfix_str}{j}'].astype(int)
  return dfsrc



import json
import pandas as pd
import numpy as np

def get_last_two_bars_OLD(df:pd.DataFrame)->{dict,dict}:
  current_bar = df.iloc[-1].copy()
  last_bar_completed = df.iloc[-2:].copy()
  completed_bars_dict = last_bar_completed.to_dict()
  current_bar_dict = current_bar.to_dict()
  return completed_bars_dict,current_bar_dict
  
def get_last_two_bars(df:pd.DataFrame):
  completed_bars=get_bar_at_index(df,-2)
  current_bar=get_bar_at_index(df,-1)
  return completed_bars,current_bar
  

def get_bar_at_index(df:pd.DataFrame,idx=-1,):
  tbar = df.iloc[idx]
  tbar_json_str = tbar.to_json()
  return json.loads(tbar_json_str)

def get_bar_at_index_v2(df:pd.DataFrame,idx=-1,format='%Y-%m-%d %H:%M:%S'):
  if 'Date' in df.columns:
    df2 = df.copy()
  else:
      df2 = df.copy().reset_index()
  df2['Date'] = pd.to_datetime(df2['Date'], format=format)
  tbar = df2.iloc[idx]
  tbar_json_str = tbar.to_json()
  return json.loads(tbar_json_str)